use std::{
    io::Read,
    ops::Deref,
    path::{Path, PathBuf},
    sync::{Arc, Mutex},
};

use ::semver::Version;
use anyhow::Context;
use clap::{Args, Parser, Subcommand};
use serde::{de::DeserializeOwned, Deserialize, Serialize};

use crate::{
    cliff::{self, changelog_parser},
    environment::get_from_environment,
    git_client::VcsClient,
    gitea_client::GiteaClient,
    ui::{ConsoleUi, DynUi},
    versioning::{next_version::NextVersion, semver::get_most_significant_version},
};

type StdinFn = Option<Arc<Mutex<dyn Fn() -> anyhow::Result<String> + Send + Sync + 'static>>>;

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
pub struct Command {
    #[command(flatten)]
    global: GlobalArgs,

    #[command(subcommand)]
    commands: Option<Commands>,

    #[clap(skip)]
    ui: DynUi,

    #[clap(skip)]
    stdin: StdinFn,
}

#[derive(Args)]
struct GlobalArgs {
    /// token is the personal access token from gitea.
    #[arg(
        env = "CUDDLE_PLEASE_TOKEN",
        long,
        long_help = "token is the personal access token from gitea. It requires at least repository write access,  it isn't required by default, but for most usecases the flow will fail without it",
        global = true,
        help_heading = "Global"
    )]
    token: Option<String>,

    /// Which repository to publish against. If not supplied remote url will be inferred from environment or fail if not present.
    #[arg(long, global = true, help_heading = "Global")]
    api_url: Option<String>,

    /// repo is the name of repository you want to release for
    #[arg(long, global = true, help_heading = "Global")]
    repo: Option<String>,

    /// owner is the name of user from which the repository belongs <user>/<repo>
    #[arg(long, global = true, help_heading = "Global")]
    owner: Option<String>,

    /// which source directory to use, if not set `std::env::current_dir` is used instead.
    #[arg(long, global = true, help_heading = "Global")]
    source: Option<PathBuf>,

    /// which branch is being run from
    #[arg(long, global = true, help_heading = "Global")]
    branch: Option<String>,

    /// whether to run in dry run mode (i.e. no pushes or releases)
    #[arg(long, global = true, help_heading = "Global")]
    dry_run: bool,

    /// Inject configuration from stdin
    #[arg(
        long,
        global = true,
        help_heading = "Global",
        long_help = "inject via stdin
cat <<EOF | cuddle-please --config-stdin 
something
something
something
EOF
config-stdin will consume stdin until the channel is closed via. EOF"
    )]
    config_stdin: bool,
}

impl Default for Command {
    fn default() -> Self {
        Self::new()
    }
}

impl Command {
    pub fn new() -> Self {
        let args = std::env::args();

        Self::new_from_args_with_stdin(Some(ConsoleUi::default()), args, || {
            let mut input = String::new();
            std::io::stdin().read_to_string(&mut input)?;
            Ok(input)
        })
    }

    pub fn new_from_args<I, T, UIF>(ui: Option<UIF>, i: I) -> Self
    where
        I: IntoIterator<Item = T>,
        T: Into<std::ffi::OsString> + Clone,
        UIF: Into<DynUi>,
    {
        let mut s = Self::parse_from(i);

        if let Some(ui) = ui {
            s.ui = ui.into();
        }

        s
    }

    pub fn new_from_args_with_stdin<I, T, F, UIF>(ui: Option<UIF>, i: I, input: F) -> Self
    where
        I: IntoIterator<Item = T>,
        T: Into<std::ffi::OsString> + Clone,
        F: Fn() -> anyhow::Result<String> + Send + Sync + 'static,
        UIF: Into<DynUi>,
    {
        let mut s = Self::parse_from(i);

        if let Some(ui) = ui {
            s.ui = ui.into();
        }
        s.stdin = Some(Arc::new(Mutex::new(input)));

        s
    }

    fn get_config(
        &self,
        current_dir: &Path,
        stdin: Option<String>,
    ) -> anyhow::Result<PleaseConfig> {
        let mut config = get_config(current_dir, stdin)?;

        self.get_from_environment(&mut config)?;

        Ok(config)
    }

    pub fn execute(self, current_dir: Option<&Path>) -> anyhow::Result<()> {
        // 1. Parse the current directory
        let current_dir = get_current_path(current_dir, self.global.source.clone())?;
        let stdin = if self.global.config_stdin {
            if let Some(stdin_fn) = self.stdin.clone() {
                let output = (stdin_fn.lock().unwrap().deref())();
                Some(output.unwrap())
            } else {
                None
            }
        } else {
            None
        };

        match &self.commands {
            Some(Commands::Release {}) => {
                tracing::debug!("running bare command");
                // 2. Parse the cuddle.please.yaml let cuddle.please.yaml take precedence
                // 2a. if not existing use default.
                // 2b. if not in a git repo abort. (unless --no-vcs is turned added)
                let _config = self.get_config(&current_dir, stdin)?;

                let owner = self.global.owner.as_ref().expect("owner to be set");
                let repo = self.global.repo.as_ref().expect("repo to be set");
                let branch = self.global.branch.as_ref().expect("branch to be set");

                let git_client = self.get_git(&current_dir)?;

                // 3. Create gitea client and do a health check
                let gitea_client = self.get_gitea_client();
                gitea_client
                    .connect(owner, repo)
                    .context("failed to connect to gitea repository")?;
                // 4. Fetch git tags for the current repository
                let tags = gitea_client.get_tags(owner, repo)?;

                let significant_tag = get_most_significant_version(tags.iter().collect());

                // 5. Fetch git commits since last git tag
                let commits = gitea_client.get_commits_since(
                    owner,
                    repo,
                    significant_tag.map(|st| st.commit.sha.clone()),
                    branch,
                )?;

                // 7. Create a versioning client
                let current_version = significant_tag
                    .map(|st| Version::try_from(st).unwrap())
                    .unwrap_or(Version::new(0, 1, 0));

                // 8. Parse conventional commits and determine next version

                let commit_strs = commits
                    .iter()
                    .map(|c| c.commit.message.as_str())
                    .collect::<Vec<&str>>();

                if commit_strs.is_empty() {
                    tracing::info!("no commits to base release on");
                    return Ok(());
                }

                let next_version = current_version.next(&commit_strs);

                // Compose changelog
                let builder =
                    cliff::ChangeLogBuilder::new(&commit_strs, next_version.to_string()).build();

                let changelog_placement = self
                    .global
                    .source
                    .as_ref()
                    .map(|s| s.join("CHANGELOG.md"))
                    .unwrap_or(PathBuf::from("CHANGELOG.md"));

                let changelog = match std::fs::read_to_string(&changelog_placement).ok() {
                    Some(existing_changelog) => builder.prepend(existing_changelog)?,
                    None => builder.generate()?,
                };

                let changelog_last_changes = changelog_parser::last_changes(&changelog)?;

                // 9b. check for release commit and release, if release exists continue
                // 10b. create release
                if let Some(first_commit) = commit_strs.first() {
                    if first_commit.contains("chore(release): ") {
                        if !self.global.dry_run {
                            gitea_client.create_release(
                                owner,
                                repo,
                                next_version.to_string(),
                                changelog_last_changes.unwrap(),
                                !next_version.pre.is_empty(),
                            )?;
                        } else {
                            tracing::debug!("creating release (dry_run)");
                        }

                        return Ok(());
                    }
                }

                // 9a. Create / Update Pr
                // Create or update branch
                git_client.checkout_branch()?;

                std::fs::write(changelog_placement, changelog.as_bytes())?;

                git_client.commit_and_push(next_version.to_string(), self.global.dry_run)?;

                let _pr_number = match gitea_client.get_pull_request(owner, repo)? {
                    Some(existing_pr) => {
                        if !self.global.dry_run {
                            gitea_client.update_pull_request(
                                owner,
                                repo,
                                next_version.to_string(),
                                changelog_last_changes.unwrap(),
                                existing_pr,
                            )?
                        } else {
                            tracing::debug!("updating pull request (dry_run)");
                            1
                        }
                    }
                    None => {
                        if !self.global.dry_run {
                            gitea_client.create_pull_request(
                                owner,
                                repo,
                                next_version.to_string(),
                                changelog,
                                self.global.branch.clone().unwrap(),
                            )?
                        } else {
                            tracing::debug!("creating pull request (dry_run)");
                            1
                        }
                    }
                };
            }

            Some(Commands::Config { command }) => match command {
                ConfigCommand::List { .. } => {
                    tracing::debug!("running command: config list");
                    let _config = self.get_config(current_dir.as_path(), stdin)?;

                    self.ui.write_str_ln("cuddle-config");
                }
            },
            Some(Commands::Gitea { command }) => {
                let git_url = url::Url::parse(&self.global.api_url.unwrap())?;

                let mut url = String::new();
                url.push_str(git_url.scheme());
                url.push_str("://");
                url.push_str(&git_url.host().unwrap().to_string());
                if let Some(port) = git_url.port() {
                    url.push_str(format!(":{port}").as_str());
                }

                let client = GiteaClient::new(url, self.global.token);
                match command {
                    GiteaCommand::Connect {} => {
                        client.connect(self.global.owner.unwrap(), self.global.repo.unwrap())?;
                        self.ui.write_str_ln("connected succesfully go gitea");
                    }
                    GiteaCommand::Tags { command } => match command {
                        Some(GiteaTagsCommand::MostSignificant {}) => {
                            let tags = client
                                .get_tags(self.global.owner.unwrap(), self.global.repo.unwrap())?;

                            match get_most_significant_version(tags.iter().collect()) {
                                Some(tag) => {
                                    self.ui.write_str_ln(&format!(
                                        "found most significant tags: {}",
                                        tag.name
                                    ));
                                }
                                None => {
                                    self.ui.write_str_ln("found no tags with versioning schema");
                                }
                            }
                        }
                        None => {
                            let tags = client
                                .get_tags(self.global.owner.unwrap(), self.global.repo.unwrap())?;
                            self.ui.write_str_ln("got tags from gitea");
                            for tag in tags {
                                self.ui.write_str_ln(&format!("- {}", tag.name))
                            }
                        }
                    },
                    GiteaCommand::SinceCommit { sha, branch } => {
                        let commits = client.get_commits_since(
                            self.global.owner.unwrap(),
                            self.global.repo.unwrap(),
                            Some(sha),
                            branch,
                        )?;
                        self.ui.write_str_ln("got commits from gitea");
                        for commit in commits {
                            self.ui.write_str_ln(&format!("- {}", commit.get_title()))
                        }
                    }
                    GiteaCommand::CheckPr {} => {
                        let pr = client.get_pull_request(
                            self.global.owner.unwrap(),
                            self.global.repo.unwrap(),
                        )?;

                        match pr {
                            Some(index) => {
                                self.ui.write_str_ln(&format!(
                                    "found cuddle-please (index={}) pr from gitea",
                                    index
                                ));
                            }
                            None => {
                                self.ui.write_str_ln("found no cuddle-please pr from gitea");
                            }
                        }
                    }
                }
            }
            Some(Commands::Doctor {}) => {
                match std::process::Command::new("git").arg("-v").output() {
                    Ok(o) => {
                        let stdout = std::str::from_utf8(&o.stdout).unwrap_or("");
                        self.ui.write_str_ln(&format!("OK: {}", stdout));
                    }
                    Err(e) => {
                        self.ui
                            .write_str_ln(&format!("WARNING: git is not installed: {}", e));
                    }
                }
            }
            None => {}
        }

        Ok(())
    }

    fn get_git(&self, current_dir: &Path) -> anyhow::Result<VcsClient> {
        VcsClient::new_git(current_dir)
    }

    fn get_from_environment(&self, config: &mut PleaseConfig) -> anyhow::Result<()> {
        let input_config = get_from_environment();
        config.merge_mut(input_config);

        Ok(())
    }

    fn get_gitea_client(&self) -> GiteaClient {
        GiteaClient::new(
            self.global.api_url.clone().expect("api_url to be set"),
            self.global.token.clone(),
        )
    }
}

#[derive(Debug, Clone, Subcommand)]
enum Commands {
    /// Config is mostly used for debugging the final config output
    Release {},

    #[command(hide = true)]
    Config {
        #[command(subcommand)]
        command: ConfigCommand,
    },

    #[command(hide = true)]
    Gitea {
        #[command(subcommand)]
        command: GiteaCommand,
    },
    /// Helps you identify missing things from your execution environment for cuddle-please to function as intended
    Doctor {},
}

#[derive(Subcommand, Debug, Clone)]
enum ConfigCommand {
    /// List will list the final configuration
    List {},
}

#[derive(Subcommand, Debug, Clone)]
enum GiteaCommand {
    Connect {},
    Tags {
        #[command(subcommand)]
        command: Option<GiteaTagsCommand>,
    },
    SinceCommit {
        #[arg(long)]
        sha: String,

        #[arg(long)]
        branch: String,
    },
    CheckPr {},
}

#[derive(Subcommand, Debug, Clone)]
enum GiteaTagsCommand {
    MostSignificant {},
}

fn get_current_path(
    optional_current_dir: Option<&Path>,
    optional_source_path: Option<PathBuf>,
) -> anyhow::Result<PathBuf> {
    let path = optional_source_path
        .or_else(|| optional_current_dir.map(|p| p.to_path_buf())) // fall back on current env from environment
        .filter(|v| v.to_string_lossy() != "") // make sure we don't get empty values
        //.and_then(|p| p.canonicalize().ok()) // Make sure we get the absolute path
        .context("could not find current dir, pass --source as a replacement")?;

    if !path.exists() {
        anyhow::bail!("path doesn't exist {}", path.display());
    }

    Ok(path)
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PleaseProjectConfig {
    pub owner: Option<String>,
    pub repository: Option<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PleaseSettingsConfig {
    pub api_url: Option<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize, Default)]
pub struct PleaseConfig {
    pub project: Option<PleaseProjectConfig>,
    pub settings: Option<PleaseSettingsConfig>,
}

impl PleaseConfig {
    fn merge(self, _config: PleaseConfig) -> Self {
        self
    }

    fn merge_mut(&mut self, _config: PleaseConfig) -> &mut Self {
        self
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct CuddleEmbeddedPleaseConfig {
    please: PleaseConfig,
}

impl From<CuddleEmbeddedPleaseConfig> for PleaseConfig {
    fn from(value: CuddleEmbeddedPleaseConfig) -> Self {
        value.please
    }
}
#[derive(Debug, Clone, Serialize, Deserialize)]
struct CuddlePleaseConfig {
    #[serde(flatten)]
    please: PleaseConfig,
}
impl From<CuddlePleaseConfig> for PleaseConfig {
    fn from(value: CuddlePleaseConfig) -> Self {
        value.please
    }
}

const CUDDLE_FILE_NAME: &str = "cuddle";
const CUDDLE_CONFIG_FILE_NAME: &str = "cuddle.please";
const YAML_EXTENSION: &str = "yaml";

fn get_config(current_dir: &Path, stdin: Option<String>) -> anyhow::Result<PleaseConfig> {
    let current_cuddle_path = current_dir
        .clone()
        .join(format!("{CUDDLE_FILE_NAME}.{YAML_EXTENSION}"));
    let current_cuddle_config_path = current_dir
        .clone()
        .join(format!("{CUDDLE_CONFIG_FILE_NAME}.{YAML_EXTENSION}"));
    let mut please_config = PleaseConfig::default();

    if let Some(config) = get_config_from_file::<CuddleEmbeddedPleaseConfig>(current_cuddle_path) {
        please_config = please_config.merge(config);
    }

    if let Some(config) = get_config_from_file::<CuddlePleaseConfig>(current_cuddle_config_path) {
        please_config = please_config.merge(config);
    }

    if let Some(input_config) = get_config_from_stdin::<CuddlePleaseConfig>(stdin.as_ref()) {
        please_config = please_config.merge(input_config);
    }

    Ok(please_config)
}

fn get_config_from_file<T>(current_cuddle_path: PathBuf) -> Option<PleaseConfig>
where
    T: DeserializeOwned,
    T: Into<PleaseConfig>,
{
    match std::fs::File::open(&current_cuddle_path) {
        Ok(file) => match serde_yaml::from_reader::<_, T>(file) {
            Ok(config) => {
                return Some(config.into());
            }
            Err(e) => {
                tracing::debug!(
                    "{} doesn't contain a valid please config: {}",
                    &current_cuddle_path.display(),
                    e
                );
            }
        },
        Err(e) => {
            tracing::debug!(
                "did not find or was not allowed to read {}, error: {}",
                &current_cuddle_path.display(),
                e,
            );
        }
    }

    None
}

fn get_config_from_stdin<'d, T>(stdin: Option<&'d String>) -> Option<PleaseConfig>
where
    T: Deserialize<'d>,
    T: Into<PleaseConfig>,
{
    match stdin {
        Some(content) => match serde_yaml::from_str::<'d, T>(content) {
            Ok(config) => {
                return Some(config.into());
            }
            Err(e) => {
                tracing::debug!("stdin doesn't contain a valid please config: {}", e);
            }
        },
        None => {
            tracing::trace!("Stdin was not set continueing",);
        }
    }

    None
}
